from .base import Model

__all__ = [
    "MatchMaker",
]


class MatchMaker(Model):
    """An implementation of the MatchMaker model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/23
    """



from .base import Model

__all__ = [
    "DeepDDS",
]


class DeepDDS(Model):
    """An implementation of the DeepDDS model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/19
    """



from .base import Model

__all__ = [
    "DeepCCI",
]


class DeepCCI(Model):
    """An implementation of the DeepCCI model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/1
    """



import math
import torch
import pandas as pd
import numpy as np
from typing import List
from torchdrug.data import PackedGraph
from chemicalx.data import LabeledTriples, DrugFeatureSet, ContextFeatureSet, DrugPairBatch


class BatchGenerator:
    """
    Generator to create batches of drug pairs efficiently.

    Args:
        batch_size (int): Number of drug pairs per batch.
        context_features (bool): Indicator whether the batch should include biological context features.
        drug_features (bool): Indicator whether the batch should include drug features.
        drug_molecules (bool): Indicator whether the batch should include drug molecules
        labels (bool): Indicator whether the batch should include drug pair labels.
    """

    def __init__(
        self,
        batch_size: int,
        context_features: bool,
        drug_features: bool,
        drug_molecules: bool,
        labels: bool,
    ):
        self.batch_size = batch_size
        self.context_features = context_features
        self.drug_features = drug_features
        self.drug_molecules = drug_molecules
        self.labels = labels

    def set_context_feature_set(self, context_feature_set: None):
        """
        Method to set the context feature set.

        Args:
            context_feature_set (ContextFeatureSet): A context feature set for feature generation.
        """
        self.context_feature_set = context_feature_set

    def set_drug_feature_set(self, drug_feature_set: None):
        """
        Method to set the drug feature set.

        Args:
            drug_feature_set (DrugFeatureSet): A drug feature set for feature generation.
        """
        self.drug_feature_set = drug_feature_set

    def set_labeled_triples(self, labeled_triples: None):
        """
        Method to set the labeled triples.

        Args:
            labeled_triples (LabeledTriples): A labeled triples object used to generate batches.
        """
        self.labeled_triples = labeled_triples

    def set_data(self, context_feature_set: None, drug_feature_set: None, labeled_triples: None):
        """
        Method to set the feature sets and the labeled triples in one pass.

        Args:
            context_feature_set (ContextFeatureSet): A context feature set for feature generation.
            drug_feature_set (DrugFeatureSet): A drug feature set for feature generation.
            labeled_triples (LabeledTriples): A labeled triples object used to generate batches.
        """
        self.set_context_feature_set(context_feature_set)
        self.set_drug_feature_set(drug_feature_set)
        self.set_labeled_triples(labeled_triples)

    def _get_context_features(self, context_identifiers: List[str]):
        """
        Getting the context features as a matrix.

        Args:
            context_identifiers (pd.Series): The context identifiers of interest.
        Returns:
            context_features (torch.FloatTensor): The matrix of biological context features.
        """
        context_features = None
        if self.context_features:
            context_features = self.context_feature_set.get_feature_matrix(context_identifiers)
        return context_features

    def _get_drug_features(self, drug_identifiers: List[str]):
        """
        Getting the global drug features as a matrix.

        Args:
            drug_identifiers (pd.Series): The drug identifiers of interest.
        Returns:
            drug_features (torch.FloatTensor): The matrix of drug features.
        """
        drug_features = None
        if self.drug_features:
            drug_features = self.drug_feature_set.get_feature_matrix(drug_identifiers)
        return drug_features

    def _get_drug_molecules(self, drug_identifiers: pd.Series) -> PackedGraph:
        """
        Getting the molecular structure of drugs.

        Args:
            drug_identifiers (pd.Series): The drug identifiers of interest.
        Returns:
            molecules (torch.PackedGraph): The molecules diagonnaly batched together for message passing.
        """
        molecules = None
        if self.drug_molecules:
            molecules = self.drug_feature_set.get_molecules(drug_identifiers)
        return molecules

    def _transform_labels(self, labels: List):
        """
        Transforming the labels from a chunk of the labeled triples frame.

        Args:
            labels (pd.Series): The drug pair binary labels.
        Returns:
            labels (torch.FloatTensor): The label target vector as a column vector.
        """
        if self.labels:
            labels = torch.FloatTensor(np.array(labels).reshape(-1, 1))
        else:
            labels = None
        return labels

    def generate_batch(self, batch_frame: pd.DataFrame):
        """
        Generate a batch of drug features, molecules, context features and labels for a set of pairs.

        Args:
            batch_frame (pd.DataFrame): The labeled pairs of interest.
        Returns:
            batch (DrugPairBatch): A batch of tensors for the pairs.
        """

        drug_features_left = self._get_drug_features(batch_frame["drug_1"])
        drug_molecules_left = self._get_drug_molecules(batch_frame["drug_1"])

        drug_features_right = self._get_drug_features(batch_frame["drug_2"])
        drug_molecules_right = self._get_drug_molecules(batch_frame["drug_2"])

        context_features = self._get_context_features(batch_frame["context"])

        labels = self._transform_labels(batch_frame["label"])

        batch = DrugPairBatch(
            batch_frame,
            drug_features_left,
            drug_molecules_left,
            drug_features_right,
            drug_molecules_right,
            context_features,
            labels,
        )

        return batch

    def __iter__(self):
        """
        Start of iteration method - shuffles the triples and resets the interator index.
        """
        self.labeled_triples.data = self.labeled_triples.data.sample(frac=1.0)
        self.sample_count = self.labeled_triples.data.shape[0]
        self.lower_frame_index = 0
        return self

    def __len__(self):
        """
        Method to set the maximal index of batches - helps tools like tqdm.
        """
        return math.ceil(self.labeled_triples.data.shape[0] / self.batch_size)

    def __next__(self):
        """
        Getting the next batch from the generator.
        """
        if self.lower_frame_index < self.sample_count:
            self.upper_frame_index = self.lower_frame_index + self.batch_size
            sub_frame = self.labeled_triples.data.iloc[
                self.lower_frame_index : min(self.upper_frame_index, self.sample_count)
            ]
            batch = self.generate_batch(sub_frame)
            self.lower_frame_index = self.upper_frame_index
            return batch
        else:
            self.frame_index = 0
            raise StopIteration



import torch
import torchdrug
import numpy as np
from typing import List, Dict, Union
from torchdrug.data import Molecule, Graph, PackedGraph


class DrugFeatureSet(dict):
    """
    Drug feature set for compunds.
    """

    def __setitem__(self, drug: str, features: Dict[str, Union[str, torch.FloatTensor]]):
        """Setting the features for a compound key

        Args:
            drug (str): Drug identifier.
            features (dict): Dictionary of smiles string and molecular features.
        """
        self.__dict__[drug] = {}
        self.__dict__[drug]["features"] = torch.FloatTensor(features["features"])
        self.__dict__[drug]["molecule"] = Molecule.from_smiles(features["smiles"])

    def __getitem__(self, drug: str):
        """Getting the features for a drug key.

        Args:
            drug (str): Drug identifier.
        Returns:
            dict: The drug features corresponding to the key.
        """
        return self.__dict__[drug]

    def __len__(self):
        """Getting the number of drugs.

        Returns:
            int: The number of drugs.
        """
        return len(self.__dict__)

    def __delitem__(self, drug: str):
        """Deleting the features for a drug key.

        Args:
            drug (str): Drug identifier.
        """
        del self.__dict__[drug]

    def clear(self):
        """Deleting all of the drugs from the drug feature set.

        Returns:
            DrugFeatureSet: An empty drug feature set.
        """
        return self.__dict__.clear()

    def has_drug(self, drug: str):
        """Checking whether a drug feature set contains a drug.

        Args:
            drug (str): Drug identifier.
        Returns:
            bool: Boolean describing whether the drug is in the drug set.
        """
        return drug in self.__dict__

    def update(self, data: Dict[str, Dict]):
        """Adding a dictionary of drug keys - feature dictionaries to a drug set.

        Args:
            data (dict): A dictionary of drug keys with feature dictionaries.
        Returns:
            DrugFeatureSet: The updated drug feature set.
        """
        return self.__dict__.update(
            {
                drug: {
                    "features": torch.FloatTensor(features["features"]),
                    "molecule": Molecule.from_smiles(features["smiles"]),
                }
                for drug, features in data.items()
            }
        )

    def keys(self):
        """Retrieving the drugs in a feature set.

        Returns:
            list: An iterator of drug identifiers.
        """
        return self.__dict__.keys()

    def values(self):
        """Retrieving the iterator of drug features.

        Returns:
            list: Feature iterator.
        """
        return self.__dict__.values()

    def items(self):
        """Retrieving the iterator of tuples containing drug identifier - feature pairs.

        Returns:
            list: An iterator of (drug - feature dictionary) tuples.
        """
        return self.__dict__.items()

    def __contains__(self, drug: str):
        """A data class method which allows the use of the 'in' operator.

        Args:
            drug (str): A drug identifier.
        Returns:
            bool: An indicator whether the drug is in the drug feature set.
        """
        return drug in self.__dict__

    def __iter__(self):
        """A data class method which allows iteration over the drug feature set.

        Returns:
            iterable: An iterable of the drug feature set.
        """
        return iter(self.__dict__)

    def get_drug_count(self) -> int:
        """Getting the number of drugs.

        Returns:
            int: The number of drugs.
        """
        return len(self.__dict__)

    def get_feature_matrix(self, drugs: List[str]) -> torch.FloatTensor:
        """Getting the drug feature matrix for a list of drugs.

        Args:
            drugs (list): A list of drug identifiers.
        Return:
            features (torch.FloatTensor): A matrix of drug features.
        """
        features = torch.cat([self.__dict__[drug]["features"] for drug in drugs])
        return features

    def get_molecules(self, drugs: List[str]) -> PackedGraph:
        """Getting the molecular structures.

        Args:
            drugs (list): A list of drug identifiers.
        Return:
            molecules (torch.PackedGraph): The molecules batched together for message passing.
        """
        molecules = Graph.pack([self.__dict__[drug]["molecule"] for drug in drugs])
        return molecules



from chemicalx.data import contextfeatureset, drugfeatureset, labeledtriples, datasetloader  # noqa:F401,F403

from chemicalx.models import (  # noqa:F401,F403
    audnnsynergy,
    caster,
    deepcci,
    deepddi,
    deepdds,
    deepdrug,
    deepsynergy,
    dpddi,
    epgcnds,
    gcnbmp,
    matchmaker,
    mhcaddi,
    mrgnn,
    ssiddi,
)

from chemicalx.version import __version__



from .base import Model

__all__ = [
    "DeepDrug",
]


class DeepDrug(Model):
    """An implementation of the DeepDrug model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/14
    """



import io
import json
import numpy as np
import pandas as pd
import urllib.request
from typing import Dict
from chemicalx.data import DrugFeatureSet, ContextFeatureSet, LabeledTriples


class DatasetLoader:
    """
    General dataset loader for the integrated drug pair scoring datasets.
    """

    def __init__(self, dataset_name: str):
        """
        Args:
            dataset_name (str): The name of the dataset.
        """
        self.base_url = "https://raw.githubusercontent.com/AstraZeneca/chemicalx/main/dataset"
        self.dataset_name = dataset_name
        assert dataset_name in ["drugcombdb", "drugcomb"]

    def generate_path(self, file_name: str) -> str:
        """
        Generating a complete url for a dataset file.

        Args:
            file_name (str): Name of the data file.
        Returns:
            data_path (str): The complete url to the dataset.
        """
        data_path = "/".join([self.base_url, self.dataset_name, file_name])
        return data_path

    def load_raw_json_data(self, path: str) -> Dict:
        """
        Given a path reading the raw JSON dataset.

        Args:
            path (str): The path to the JSON file.
        Returns:
            raw_data (dict): A dictionary with the data.
        """
        with urllib.request.urlopen(path) as url:
            raw_data = json.loads(url.read().decode())
        return raw_data

    def load_raw_csv_data(self, path: str) -> pd.DataFrame:
        """
        Reading the labeled triples CSV in memory.

        Args:
            path (str): The path to the triples CSV file.
        Returns:
            raw_data (pd.DataFrame): A pandas DataFrame with the data.
        """
        data_bytes = urllib.request.urlopen(path).read()
        types = {"drug_1": str, "drug_2": str, "context": str, "label": float}
        raw_data = pd.read_csv(io.BytesIO(data_bytes), encoding="utf8", sep=",", dtype=types)
        return raw_data

    def get_context_features(self):
        """
        Reading the context feature set.

        Returns:
            context_feature_set (ContextFeatureSet): The ContextFeatureSet of the dataset of interest.
        """
        path = self.generate_path("context_set.json")
        raw_data = self.load_raw_json_data(path)
        raw_data = {k: np.array(v).reshape(1, -1) for k, v in raw_data.items()}
        context_feature_set = ContextFeatureSet()
        context_feature_set.update(raw_data)
        return context_feature_set

    def get_drug_features(self):
        """
        Reading the drug feature set.

        Returns:
            drug_feature_set (DrugFeatureSet): The DrugFeatureSet of the dataset of interest.
        """
        path = self.generate_path("drug_set.json")
        raw_data = self.load_raw_json_data(path)
        raw_data = {
            k: {"smiles": v["smiles"], "features": np.array(v["features"]).reshape(1, -1)} for k, v in raw_data.items()
        }
        drug_feature_set = DrugFeatureSet()
        drug_feature_set.update(raw_data)
        return drug_feature_set

    def get_labeled_triples(self):
        """
        Getting the labeled triples file from the storage.

        Returns:
            labeled_triples (LabeledTriples): The labeled triples in the dataset.
        """
        path = self.generate_path("labeled_triples.csv")
        raw_data = self.load_raw_csv_data(path)
        labeled_triples = LabeledTriples()
        labeled_triples.update_from_pandas(raw_data)
        return labeled_triples



from .contextfeatureset import *  # noqa:F401,F403
from .drugfeatureset import *  # noqa:F401,F403
from .labeledtriples import *  # noqa:F401,F403
from .datasetloader import *  # noqa:F401,F403
from .drugpairbatch import *  # noqa:F401,F403
from .batchgenerator import *  # noqa:F401,F403



class DrugPairBatch:
    """
    A data class to store a labeled drug pair batch.

    Args:
        identifiers (pd.DataFrame or None): A dataframe with drug pair, context and label columns.
        drug_features_left (torch.FloatTensor or None): A matrix of molecular features for the left hand drugs.
        drug_molecules_left (torchdrug.PackedGraph or None): Packed molecules for the left hand drugs.
        drug_features_right (torch.FloatTensor or None): A matrix of molecular features for the right hand drugs.
        drug_molecules_right (torchdrug.PackedGraph or None): Packed molecules for the right hand drugs.
        context_features (torch.FloatTensor or None): A matrix of biological/chemical context features.
        labels (torch.FloatTensor or None): A vector of drug pair labels.
    """

    def __init__(
        self,
        identifiers=None,
        drug_features_left=None,
        drug_molecules_left=None,
        drug_features_right=None,
        drug_molecules_right=None,
        context_features=None,
        labels=None,
    ):
        self.identifiers = identifiers
        self.drug_features_left = drug_features_left
        self.drug_molecules_left = drug_molecules_left
        self.drug_features_right = drug_features_right
        self.drug_molecules_right = drug_molecules_right
        self.context_features = context_features
        self.labels = labels



from .base import Model

__all__ = [
    "SSIDDI",
]


class SSIDDI(Model):
    """An implementation of the SSI-DDI model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/11
    """



import pandas as pd
from typing import List, Tuple
from sklearn.model_selection import train_test_split


class LabeledTriples:
    """
    Labeled triples for drug pair scoring.
    """

    def __init__(self):
        self.columns = ["drug_1", "drug_2", "context", "label"]
        self.types = {"drug_1": str, "drug_2": str, "context": str, "label": float}
        self.data = pd.DataFrame(columns=self.columns).astype(self.types)

    def drop_duplicates(self):
        """
        Dropping the duplicated entries.
        """
        self.data = self.data.drop_duplicates()

    def update_from_pandas(self, data: pd.DataFrame):
        """
        Updating the labeled triples from a dataframe.

        Args:
            data (pd.DataFrame): A dataframe of labeled triples.
        """
        self.data = pd.concat([self.data, data])

    def update_from_list(self, data: List[List]):
        """
        Updating the labeled triples from a list.

        Args:
            data (list): A list of labeled triples.
        """
        data = pd.DataFrame(data, columns=self.columns)
        self.data = pd.concat([self.data, data])

    def __add__(self, value):
        """
        Adding together the triples in two LabeledTriples objects - syntactic sugar for '+'.

        Args:
            value (LabeledTriples): Another LabeledTriples object for the addition.
        Returns:
            new_triples (LabeledTriples): A LabeledTriples object after the addition.
        """
        new_triples = LabeledTriples()
        new_triples.update_from_pandas(pd.concat([self.data, value.data]))
        return new_triples

    def get_drug_count(self) -> int:
        """
        Getting the number of drugs in the labeled triples dataset.

        Returns
            int: The number of unique compounds in the labeled triples dataset.
        """
        return pd.unique(self.data[["drug_1", "drug_2"]].values.ravel("K")).shape[0]

    def get_context_count(self) -> int:
        """
        Getting the number of unique contexts in the labeled triples dataset.

        Returns
            int: The number of unique contexts in the labeled triples dataset.
        """
        return self.data["context"].nunique()

    def get_combination_count(self) -> int:
        """
        Getting the number of unique drug pairs in the labeled triples dataset.

        Returns
            int: The number of unique pairs in the labeled triples dataset.
        """
        combination_count = self.data[["drug_1", "drug_2"]].drop_duplicates().shape[0]
        return combination_count

    def get_labeled_triple_count(self) -> int:
        """
        Getting the number of triples in the labeled triples dataset.

        Returns
            int: The number of triples in the labeled triples dataset.
        """
        triple_count = self.data.shape[0]
        return triple_count

    def get_positive_count(self) -> int:
        """
        Getting the number of positive triples in the dataset.

        Returns
            int: The number of positive triples.
        """
        return int(self.data["label"].sum())

    def get_negative_count(self) -> int:
        """
        Getting the number of negative triples in the dataset.

        Returns
            int: The number of negative triples.
        """
        return self.get_labeled_triple_count() - self.get_positive_count()

    def get_positive_rate(self) -> float:
        """
        Getting the ratioof positive triples in the dataset.

        Returns
            float: The ratio of positive triples.
        """
        return self.data["label"].mean()

    def get_negative_rate(self) -> float:
        """
        Getting the ratio of positive triples in the dataset.

        Returns
            float: The ratio of negative triples.
        """
        return 1.0 - self.data["label"].mean()

    def train_test_split(self, train_size: float = 0.8, random_state: int = 42) -> Tuple:
        """
        Splitting the LabeledTriples object for training and testing.

        Args:
            train_size (float): The ratio of training triples. Default is 0.8.
            random_state (int): The random seed. Default is 42.
        Returns
            train_labeled_triples (LabeledTriples): The training triples.
            test_labeled_triples (LabeledTriples): The testing triples.
        """
        train_data, test_data = train_test_split(self.data, train_size=train_size, random_state=random_state)
        train_labeled_triples = LabeledTriples()
        test_labeled_triples = LabeledTriples()
        train_labeled_triples.update_from_pandas(train_data)
        test_labeled_triples.update_from_pandas(test_data)
        return train_labeled_triples, test_labeled_triples



from .base import Model

__all__ = [
    "EPGCNDS",
]


class EPGCNDS(Model):
    """An implementation of the EPGCNDS model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/22
    """



from setuptools import find_packages, setup

install_requires = [
    "numpy",
    "torch>=1.10.0",
    "torchdrug",
    "torch-scatter>=2.0.8",
    "pandas<=1.3.5",
    "scipy",
    "tqdm",
    "six",
    "scikit-learn",
    "class-resolver",
]


setup_requires = ["pytest-runner"]

tests_require = ["pytest", "pytest-cov", "mock", "unittest"]

keywords = [
    "drug",
    "deep-learning",
    "deep-chemistry",
    "deep-ai",
    "torch-drug",
    "synergy-prediction",
    "synergy",
    "drug-combination",
    "deep-synergy",
    "drug-interaction",
    "chemistry",
    "pharma",
]


setup(
    name="chemicalx",
    packages=find_packages(),
    version="0.0.7",
    license="Apache License, Version 2.0",
    description="A Deep Learning Library for Drug Pair Scoring.",
    author="Benedek Rozemberczki and Charles Hoyt",
    author_email="",
    url="https://github.com/AstraZeneca/chemicalx",
    download_url="https://github.com/AstraZeneca/chemicalx/archive/v0.0.7.tar.gz",
    keywords=keywords,
    install_requires=install_requires,
    setup_requires=setup_requires,
    tests_require=tests_require,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
    ],
)



from .base import Model

__all__ = [
    "DeepDDI",
]


class DeepDDI(Model):
    """An implementation of the DeepDDI model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/2
    """



from .base import Model

__all__ = [
    "DeepSynergy",
]


class DeepSynergy(Model):
    """An implementation of the DeepSynergy model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/16
    """



from .base import Model

__all__ = [
    "MHCADDI",
]


class MHCADDI(Model):
    """An implementation of the MHCADDI model.

    .. seealso:: https://github.com/AstraZeneca/chemicalx/issues/13
    """
